#include <stdio.h>

#include "utf_tab.h"
#include "util.h"
#include "error_stuff.h"

void analyze_CPK(FILE *infile, long file_length);

int main(int argc, char **argv)
{
    printf("cpk_unpack 0.0\n\n");
    if (argc != 2)
        fprintf(stderr,"Incorrect program usage\n\nusage: %s file\n",argv[0]);

    /* open file */
    FILE *infile = fopen(argv[1], "rb");
    CHECK_ERRNO(!infile, "fopen");

    /* get file size */
    CHECK_ERRNO(fseek(infile, 0 , SEEK_END) != 0, "fseek");
    long file_length = ftell(infile);
    CHECK_ERRNO(file_length == -1, "ftell");

    rewind(infile);

    analyze_CPK(infile, file_length);

    exit(EXIT_SUCCESS);
}

void analyze_CPK(FILE *infile, long file_length)
{
    const long CpkHeader_offset = 0x0;
    char *toc_string_table = NULL;

    /* check header */
    {
        unsigned char buf[4];

        CHECK_ERRNO(fseek(infile, CpkHeader_offset, SEEK_SET) != 0, "fseek");

        size_t bytes_read = fread(buf, 1, 4, infile);
        CHECK_FILE(bytes_read != 4, infile, "fread");

        static const char CPK_signature[4] = "CPK "; /* intentionally unterminated */
        CHECK_ERROR (memcmp(buf, CPK_signature, sizeof(CPK_signature)), "CPK signature not found");
    }

    /* get TOC offset */
    long toc_offset;
    {
        struct utf_query query;
        query.index = 0;
        query.name = "TocOffset";

        struct utf_query_result result = query_utf(infile, CpkHeader_offset+0x10, &query);

        CHECK_ERROR (!result.valid, "didn't find valid CpkHeader");
        CHECK_ERROR (!result.found, "TocOffset not found");
        CHECK_ERROR (result.type != COLUMN_TYPE_8BYTE, "TocOffset is wrong type");
        CHECK_ERROR (result.rows != 1, "wrong number of rows in CpkHeader");
        toc_offset = result.value.value_u64;
    }

    /* get content offset */
    long content_offset;
    {
        struct utf_query query;
        query.index = 0;
        query.name = "ContentOffset";

        struct utf_query_result result = query_utf(infile, CpkHeader_offset+0x10, &query);

        CHECK_ERROR (!result.found, "ContentOffset not found");
        CHECK_ERROR (result.type != COLUMN_TYPE_8BYTE, "ContentOffset is wrong type");
        content_offset = result.value.value_u64;
    }

    /* get file count from CpkHeader */
    long CpkHeader_count;
    {
        struct utf_query query;
        query.index = 0;
        query.name = "Files";

        struct utf_query_result result = query_utf(infile, CpkHeader_offset+0x10, &query);

        CHECK_ERROR (!result.found, "\"File\" not found");
        CHECK_ERROR (result.type != COLUMN_TYPE_4BYTE, "\"File\" is wrong type");
        CpkHeader_count = result.value.value_u32;
    }

    /* check TOC header */
    {
        unsigned char buf[4];

        CHECK_ERRNO(fseek(infile, toc_offset, SEEK_SET) != 0, "fseek");

        size_t bytes_read = fread(buf, 1, 4, infile);
        CHECK_FILE(bytes_read != 4, infile, "fread");

        static const char TOC_signature[4] = "TOC "; /* intentionally unterminated */
        CHECK_ERROR (memcmp(buf, TOC_signature, sizeof(TOC_signature)), "TOC signature not found");
    }

    /* get TOC entry count, string table offset */
    long toc_entries;
    long toc_string_table_offset;
    long toc_string_table_size;
    {
        struct utf_query_result result = query_utf(infile, toc_offset+0x10, NULL);

        CHECK_ERROR (!result.valid, "didn't find valid TOC");
        toc_entries = result.rows;
        toc_string_table_offset = toc_offset + 0x10 + 8 + result.string_offset;
        toc_string_table_size = result.data_offset - result.string_offset;
    }

    /* check that counts match */
    CHECK_ERROR( toc_entries != CpkHeader_count, "CpkHeader file count and TOC entry count do not match" );

    /* load string table */
    toc_string_table = malloc(toc_string_table_size + 1);
    {
        CHECK_ERRNO(!toc_string_table, "malloc");
        memset(toc_string_table, 0, toc_string_table_size+1);
        CHECK_ERRNO(fseek(infile, toc_string_table_offset, SEEK_SET) != 0, "fseek");
        size_t bytes_read = fread(toc_string_table, 1, toc_string_table_size, infile);
        CHECK_FILE(bytes_read != toc_string_table_size, infile, "fread");
    }

    /* extract files */
    for (int i = 0; i < toc_entries; i++)
    {
        /* get file name */
        const char *file_name;
        {
            struct utf_query query;
            query.index = i;
            query.name = "FileName";

            struct utf_query_result result = query_utf(infile, toc_offset+0x10, &query);

            CHECK_ERROR (!result.found, "FileName not found");
            CHECK_ERROR (result.type != COLUMN_TYPE_STRING, "FileName is wrong type");
            file_name = toc_string_table + result.value.value_string;
        }

        /* get file size */
        long file_size;
        {
            struct utf_query query;
            query.index = i;
            query.name = "FileSize";

            struct utf_query_result result = query_utf(infile, toc_offset+0x10, &query);

            CHECK_ERROR (!result.found, "FileSize not found");
            CHECK_ERROR (result.type != COLUMN_TYPE_4BYTE, "FileSize is wrong type");
            file_size = result.value.value_u32;
        }

        /* get file offset */
        long file_offset;
        {
            struct utf_query query;
            query.index = i;
            query.name = "FileOffset";

            struct utf_query_result result = query_utf(infile, toc_offset+0x10, &query);

            CHECK_ERROR (!result.found, "FileOffset not found");
            CHECK_ERROR (result.type != COLUMN_TYPE_8BYTE, "FileOffset is wrong type");
            file_offset = content_offset + result.value.value_u64;
        }

        printf("%s %lx %ld\n", file_name, (unsigned long)file_offset, file_size);

        FILE *outfile = fopen(file_name, "wb");
        CHECK_ERRNO(!outfile, "fopen");
        dump(infile, outfile, file_offset, file_size);
        CHECK_ERRNO(fclose(outfile) != 0, "fclose");
    }

    if (toc_string_table)
    {
        free(toc_string_table);
        toc_string_table = NULL;
    }
}
